=begin

LATHE TOOLS
Version 1.2.0
31st March 2013
(c) Trog Luddite (S.Munns)

Enables quick construction of lathed forms:

Usage:
- Start with a selection of...
  * A line or construction line to be the axis of rotation.
  * At least one face co-planar to the line (multiple coplanar face can be selected)
  The axis line is allowed to be an edge of one of the selected faces;
  for example, when using a face as a drawing plane for a cross-section.
- Choose 'Quick Lathe' from the 'Tools' menu.
- You will be prompted for...
  * Number of Segments - how many faces you get around the sweep of the lathe.
  * Lathe Angle (degrees) - the total sweep of the lathe.
  * Smooth Angle (degrees) - below this angle edges are smoothed.
  * Axis c-line - whether to add a construction line along the axis.
  * Entity type - choose from loose geometry, a group, or a component.
  * Subdivide Components - divide the lathed form into a number of segments
                          each of which will be created as a component instance.
  * Erase Profile? - whether or not to keep the original profile geometry.
  * Keep Coplanar Edges - or merge coplanar faces into one.
  * Name - a root name to use when making groups or components.
  * Arc End Style - when <360deg sweep whether to cap the ends, leave them open
                    or open with hidden edges (for segmented components)
  * Arc End perpendicular - terminate ends in half faces at 90deg to the axis
  
- Settings are remembered between uses (just stab 'Enter' to confirm)
- Most invalid input conditions are tested for, and you will be prompted to try again.
  
Advantages over the standard "Follow me" tool include:
- No need to construct a circle to follow around.
- Multiple faces lathed in one pass.
- The circular 'hoops' are kept intact for easy editing after lathing.
- Turning by less than full circle doesn't create end 'stubs' at the extrusion ends.
- Tests for 'inside out' forms and attempts to correctly orient the faces.
- Automatic grouping and/or marking the axis.

Known limitations and issues:
- Only recognises face outer loops; no holes!
- 'End Caps' won't stay as curves when push/pulled.

LICENCE.
You may use, copy, distribute and modify this code without any restriction except
that it remains attributed to Trog Luddite (S.Munns).
No warranty is granted or implied; the software is used entirely at the user's
own risk, and the author accepts no responsibility for any unforeseen consequences of
its use, including, but not limited to; loss of data, damage to computer systems,
loss of earnings or reputation etc.

Bugs and suggestions for improvement will be glady accepted, directed to user
'Trogluddite' on the internet forum 'SketchUcation', though no guarantee is given that
there will be future upgrades or bug fixes.

Hope you find it useful!

ChangeLog
- v1.0.0  First attempt.
- v1.0.1  Fixed z-axis misalignment of points lying on the x-axis.
- v1.0.2  Option to keep/remove coplanar edges.
- v1.1.0  More options: End Caps, Components and arrays.  Code re-factor.
- v1.1.1  Minor bugfix - code typo's.
- v1.2.0  Non-coplanar faces and negative angles now possible.
          Face winding re-jigged to give always right way faces (I hope!)

Credits:
As ever, many thanks to the guru's and users of the SkeuchUcation forum, especially...
- ThomThom, Dan Rathburn, TIG, sdmitch, Chris Fullmer - for their incredible Ruby insights.
- Box, Dave R, Pilou, Gilles - for the excellent suggestions and quality bug-reporting!

=end

require('sketchup')  # Standard 'helper' methods.

module Trog

module LatheTools  # Author namespace

  CIRCLE = Math::PI * 2

  private # Methods for internal use only


  def self.get_sel_data(sel)
  # Extracts the required data from the passed in selection
  # Ignores any faces which are not coplanar to the axis line.
  # Returns: [face_data, old_geom]
  #   face_data = [{:trans => <face transformation>, :points => <vertex positions>},...]
  #	  old_geom = [<original entities to delete>]
  
    face_data = []
    old_geom = []
    
    # Sort selection
    edges = sel.grep(Sketchup::Edge)
    faces = sel.grep(Sketchup::Face)
    clines = sel.grep(Sketchup::ConstructionLine)
    
    # Check for valid selection
    if edges.length == 1 && faces.length > 0 && clines.length == 0
      edge = edges[0]
      line = edge.line
      origin = edge.start.position
      end_pt = edge.end.position
      z_axis = line[1].normalize
    elsif edges.length == 0 && faces.length > 0 && clines.length == 1
      origin, z_axis = clines[0].position, clines[0].direction.normalize
      line = [origin, z_axis]
      end_pt = origin + z_axis
    else
      raise(ArgumentError,"需要选择一条边或构造线,和至少一个面.")
      return nil
    end
	
    # Get data for each face, removing any which are not coplanar
    faces.each do |face|
    
      # Decide if co_planar
      plane = face.plane
      planar = (origin.on_plane?(plane) && end_pt.on_plane?(plane))
    
      # If valid add all face entities to the old_geom array
      old_geom << face
      old_geom << face.edges
      
      points = face.outer_loop.vertices.map{|pt| pt.position}
      
      # Get other axes
      x_axis = nil
      points.each do |pt|
        if pt.distance_to_line(line).to_l > 0.0
          x_axis = pt.project_to_line(line).vector_to(pt).normalize
          break
        end
      end
      y_axis = z_axis * x_axis      
  
      # Get approximate transformations
      trans_to = Geom::Transformation.axes(origin,x_axis,y_axis,z_axis)
      trans_from = trans_to.inverse
      
      # Collect points
      points.each{|pt| pt.transform!(trans_from)}
      
      # Z-shift to local origin
      low_z = points.sort{|pt1,pt2| pt1.z <=> pt2.z}[0].z
      z_trans = Geom::Transformation.translation([0, 0, -low_z])
      points.each{|pt| pt.transform!(z_trans)}
      trans_to = trans_to * z_trans.inverse
      
      # Reverse if the face normal points the wrong way
      if y_axis.angle_between(face.normal) > Math::PI
        points.reverse!
      end
      
      # Add data to the array
      face_data << {:transform=>trans_to, :points=>points, :planar=>planar}
      
    end # face.each
    
    # Error if no co_planar faces were found
    raise(ArgumentError,"未发现面") if face_data.empty?
    
    # Success
    [face_data, old_geom.flatten]
  
  end # get_sel_data
  
  
  def self.make_hoops(ents, points, angle, sides, perpendicular = false)
  # Builds arc 'hoops' around the extrusion
  # Returns: [hoop_grp, hoop_points]
  
    # Hoop arcs go into a new sub_group
    hoops = []
    hoop_grp = ents.add_group
    hoop_ents = hoop_grp.entities
    
    # Make each hoop
    points.each do |pt|
      x, y, z = pt.to_a
      if x.to_l == 0
        hoops << [Geom::Point3d.new(x, y, z)]
      else
        radius = Math::hypot(x,y)
        if perpendicular
          s_angle = angle.abs
          seg_ang = s_angle / sides
          end_ang= (s_angle + seg_ang) * 0.5
          arc = hoop_ents.add_arc([0,0,z], [x,y,0], Z_AXIS, radius, -end_ang, end_ang, sides+1)
          edge_s = arc[0]
          edge_e = arc[-1]
          edge_s.split(0.5)
          hoop_ents.erase_entities(edge_s, edge_e.split(0.5))
          arc = edge_e.curve.edges
          arc_pts = arc.map{|edge| edge.start.position}
          arc_pts.push(arc[-1].end.position)
          hoops << arc_pts
        else
          if angle < 0.0
            s_angle = angle ; e_angle = 0.0
          else
            s_angle = 0.0 ; e_angle = angle
          end
          arc = hoop_ents.add_arc([0,0,z], [x,y,0], Z_AXIS, radius, s_angle, e_angle, sides)
          arc_pts = arc.map{|edge| edge.start.position}
          arc_pts.push(arc[-1].end.position)
          hoops << arc_pts
        end
      end
    end
      
    [hoop_grp, hoops]
    
  end # make_hoops
  
  
  def self.make_mesh(hoops,planar = false)
  # Uses the hoop point data to make a 3d mesh of new faces.
  # Returns: PolygonMesh object
  
    mesh = Geom::PolygonMesh.new
    hoop1 = hoops[-1]
    hoop1_len = hoop1.length
    
    hoops.each do |hoop2|
      hoop2_len = hoop2.length
      unless hoop1_len == 1 && hoop2_len == 1
        if hoop1_len == 1
          pt3 = hoop1[0]
          (0...hoop2_len-1).each do |index|
            pt1, pt2 = hoop2[index,2]
            mesh.add_polygon(pt3,pt2,pt1)
          end
        elsif hoop2_len == 1
          pt3 = hoop2[0]
          (0...hoop1_len-1).each do |index|
            pt1, pt2 = hoop1[index,2]
            mesh.add_polygon(pt1,pt2,pt3)
          end
        else
          (0...hoop1_len-1).each do |index|
            pt1a, pt2a = hoop1[index,2]
            pt1b, pt2b = hoop2[index,2]
            if planar
              mesh.add_polygon(pt1b,pt1a,pt2a,pt2b)
            else
              mesh.add_polygon(pt1b,pt1a,pt2a)
              mesh.add_polygon(pt1b,pt2a,pt2b)
            end
          end
        end
      end
      hoop1 = hoop2
      hoop1_len = hoop2_len
    end
    
    mesh

  end # make_mesh
  
  
  def self.make_ends(ents, hoops)
  # Build the two end faces as a group and return the group
  
    ends_grp = ents.add_group
    ends_ents = ends_grp.entities
    
    points = hoops.collect{|hoop| hoop[0]}
    ends_ents.add_face(points)
    
    points = hoops.collect{|hoop| hoop[-1]}
    ends_ents.add_face(points.reverse)
    
    ends_grp
    
  end # make_ends
  
  
  def self.smooth_edges(group, angle, del_coplanar, hide_ends)
  # Takes a group and smooths the edges by face normal
  # and optionally deletes co-planar edges.
  
    keep_coplanar = !del_coplanar
    coplanar_edges = []
    ents = group.entities
    edges = ents.grep(Sketchup::Edge)
    
    edges.each do |edge|
      next if edge.smooth? && keep_coplanar
      faces = edge.faces
      num_faces = faces.size
      if num_faces == 2
        f1, f2 = faces
        angle_diff = f1.normal.angle_between(f2.normal)
        if del_coplanar && angle_diff == 0.0
          coplanar_edges << edge
        elsif angle_diff <= angle
          edge.smooth = true
          edge.soft = true
        end
      elsif num_faces == 1 && hide_ends
        edge.hidden = true
      end        
    end
    
    ents.erase_entities(coplanar_edges) unless coplanar_edges.empty?
    
  end # smooth_edges


  def self.face_outwards(group,trans)
  # Tries to detect an insde out group and turn it
  # the right way around.
  
    faces = group.entities.grep(Sketchup::Face)
    reversed = nil
    mod = Sketchup.active_model
    
    faces.each do |face|
      # Find a point on the face
      verts = face.outer_loop.vertices
      pt1, pt2 = verts[0].position, verts[2].position
      test_pt = Geom::Point3d.linear_combination(0.5, pt1, 0.5, pt2)
      test_pt.transform(trans)
      # Raytest forwards and reversed
      test_vect = face.normal.transform(trans)
      front = mod.raytest(test_pt, test_vect)
      back = mod.raytest(test_pt, test_vect.reverse)
      # State is known if either ray misses the group
      if front.nil? || front[1][0] != group
        reversed = false
        break
      elsif
        back.nil? || back[1][0] != group
        reversed = true
        break
      end
    end
    
    faces.each{|face| face.reverse!} if reversed
    
  end # face_outwards
  

  public # Accessible methods for user interface

  def self.quick_lathe
  
    mod = Sketchup.active_model
    
    # Queue to undo stack (SU version aware)
    if mod.method(:start_operation).arity == 1
      mod.start_operation("Quick Lathe")
    else
      mod.start_operation("Quick Lathe",true)
    end
    
    # Get stuff from the model
    ents = mod.active_entities
    sel = mod.selection
    defs = mod.definitions
    face_data, old_geom = get_sel_data(sel)
    num_faces = face_data.length
	
    # Get settings
    prompts = ['段数','车削角度(度)','软化/平滑边线 (度)',
              "沿轴添加构造线?","实体类型","细分组件",
              "删除截面?","保留共面边线?",
              "名称","弧端类型","弧端垂直?"]
    def_sides = Sketchup.read_default("Quick Lathe","Sides",24)
    def_angle = Sketchup.read_default("Quick Lathe","Angle",360.0)
    def_smooth = Sketchup.read_default("Quick Lathe","Smooth",22.5)
    def_axis = Sketchup.read_default("Quick Lathe","Axis","Yes")
    def_type = Sketchup.read_default("Quick Lathe","Type","Group")
    def_sub = Sketchup.read_default("Quick Lathe","SubDiv",1)
    def_erase = Sketchup.read_default("Quick Lathe","Erase","Yes")
    def_coplanar = Sketchup.read_default("Quick Lathe","Coplanar","No")
    def_name = Sketchup.read_default("Quick Lathe","Name","Lathed Form")
    def_caps = Sketchup.read_default("Quick Lathe","Caps","Capped")
    def_perp = Sketchup.read_default("Quick Lathe","Perp","No")
    defaults = [def_sides, def_angle, def_smooth,def_axis,
                def_type,def_sub,def_erase,def_coplanar,def_name,def_caps,def_perp]
    choices = ["","","","Yes|No","几何图形|组|组件","",
                "Yes|No","Yes|No","","无|加盖|隐藏","Yes|No"]
    count_string = (num_faces == 1 ? "face" : "faces")
    title = "快速车削 #{num_faces} #{count_string}."
    begin
      values = UI.inputbox(prompts,defaults,choices,title)
      return nil unless values
      sides, angle, smooth, add_axis, type, sub_div, erase_old,
        keep_coplanar, name, arc_ends, perpendicular = values
      name = "Lathed Form" if name.empty?
      raise(ArgumentError,"分段必须大于2!") if sides < 2
      raise(ArgumentError,"角度必须 >0 并 <=360") if angle <= -360 || angle  > 360.0
      raise(ArgumentError,"软化/平滑必须在 0 和 180 之间") if smooth < 0.0 || smooth > 180.0
      raise(ArgumentError,"至少给1个细分") if sub_div < 1
    rescue => error
      UI.messagebox("#{error.message}, 按 OK 再试一次")
      retry
    end
    Sketchup.write_default("Quick Lathe","Sides",sides) unless sides == def_sides
    Sketchup.write_default("Quick Lathe","Angle",angle) unless angle == def_angle
    Sketchup.write_default("Quick Lathe","Smooth",smooth) unless smooth == def_smooth
    Sketchup.write_default("Quick Lathe","Axis",add_axis) unless add_axis == def_axis
    Sketchup.write_default("Quick Lathe","Type",type) unless type == def_type
    Sketchup.write_default("Quick Lathe","SubDiv",sub_div) unless sub_div == def_sub
    Sketchup.write_default("Quick Lathe","Erase",erase_old) unless erase_old == def_erase
    Sketchup.write_default("Quick Lathe","Coplanar",keep_coplanar) unless keep_coplanar == def_coplanar
    Sketchup.write_default("Quick Lathe","Name",name) unless name == def_name
    Sketchup.write_default("Quick Lathe","Caps",arc_ends) unless arc_ends == def_caps
    Sketchup.write_default("Quick Lathe","Perp",perpendicular) unless perpendicular == def_perp
    
    # Prepare arguments
    add_axis = (add_axis == "Yes")
    erase_old = (erase_old == "Yes")
    delete_coplanar = (keep_coplanar == "No")
    perpendicular = (perpendicular == "Yes")
    if type == "Component" && sub_div > 1
      angle /= sub_div
      sides = (sides.to_f / sub_div).ceil
      sides = [2,sides].sort[1]
    end
    angle *= Math::PI / 180
    smooth *= Math::PI / 180
    
    # Delete the selection if option chosen
    ents.erase_entities(old_geom) if erase_old
    
    # Lathe each face in turn
    face_data.each do |this_face|
    
      trans = this_face[:transform]
      points = this_face[:points]
      planar = this_face[:planar]
      
#      points.reverse! if angle < 0.0
    
      # Entity grouping
      if type == "Component"
        uniq_name = defs.unique_name(name)
        grp = defs.add(uniq_name)
        grp_ents = grp.entities
      else
        grp = ents.add_group
        grp.name = name
        grp_ents = grp.entities
        grp.transformation = trans
      end
    
      # Draw arcs 'hoops' in a sub-group and get point data
      hoop_grp, hoops = make_hoops(grp_ents, points, angle, sides, perpendicular)
      
      # Add the new surface
      grp_ents.add_faces_from_mesh(make_mesh(hoops,planar))
      
      # Drape the 'hoops' over the surface
      hoop_grp.explode
      
      # Add end caps if less than a full turn
      if angle.abs < CIRCLE && arc_ends == "Capped"
        ends_grp = make_ends(grp_ents,hoops) if angle.abs < CIRCLE
        ends_grp.explode
      end

      # Make things face outwards
 ####     face_outwards(grp, trans)
      
      # Smooth and delete co-planar edges
      smooth_edges(grp, smooth, delete_coplanar, arc_ends == "Hidden")
      
      # Options
      grp_ents.add_cline([0,0,0],Z_AXIS,'-.-') if add_axis
      case type
      when "Geometry"
        grp.explode
      when "Component"
        ents.add_instance(grp,trans)
        if sub_div > 1
          (1...sub_div).each do |index|
            comp_trans = trans * Geom::Transformation.rotation([0,0,0],Z_AXIS,angle*index)
            ents.add_instance(grp,comp_trans)
          end
        end
      end
   
    end # face.each
   
    # Succesful completion
    mod.commit_operation
  
  # Trap exceptions
  rescue => error
    UI.messagebox("错误: #{error.message}")
    mod.abort_operation
    raise error
  end #method quick_lathe

  # Add menu item
  unless file_loaded?('Quick Lathe.rb')
    menu = UI.menu('Tools')
    item = menu.add_item('快速车削'){self.quick_lathe}
    file_loaded('Quick Lathe.rb')
  path = Sketchup.find_support_file("Plugins/Quick Lathe")
  quickLathe_tb = UI::Toolbar.new("快速车削")
  #传递材质至面
  cmd = UI::Command.new(("快速车削")) {self.quick_lathe}
  cmd.small_icon = File.join(path,"icons/CX_24.png")
  cmd.large_icon = File.join(path,"icons/CX_16.png")
  cmd.tooltip = "快速车削"
  cmd.status_bar_text = "选中截面及一根中线，点击按钮设置参数即可对其进行车削（旋转放样）"
  quickLathe_tb.add_item(cmd) 
  quickLathe_tb.show
  end
  
end #module LatheTools

end #module Trog
